from sys import stdin

from promplate.prompt.chat import Message, assistant, system, user
from reactivity import signal
from typer import Argument, Option, Typer

from .impl import default_model, get_client
from .markdown import streaming_markdown
from .utils import get_user_message

app = Typer()


@app.command()
def ask(message: str = Argument(""), model: str = Option(default_model, "--model", "-m")):
    if not message:
        message = get_user_message()

    messages: list[Message] = [user > message]
    if not stdin.isatty():
        messages.insert(0, system > stdin.read())

    while True:
        out = signal("")
        with streaming_markdown(out.get):
            for i in get_client().generate(messages, model=model):
                out.update(lambda last: last + i)

        messages.append(assistant > out.get())

        messages.append(user > get_user_message())



from functools import cache

from m.config.load import load_config

from .types import ChatConfig

chat_config: ChatConfig = load_config()["chat"]  # type: ignore

default_model = chat_config["options"].get("model", "gpt-4o-mini")


@cache
def get_client():
    from promplate.llm.openai import SyncChatOpenAI

    return SyncChatOpenAI(api_key=chat_config.get("openai_api_key"), base_url=chat_config.get("openai_base_url"))



from asyncio import Event, Lock
from collections.abc import Callable
from contextlib import contextmanager
from threading import Thread

from prompt_toolkit import ANSI, Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import FormattedTextControl, Layout, Window
from reactivity import async_effect, derived
from rich.console import Console, RenderResult
from rich.containers import Renderables
from rich.markdown import Markdown


class TruncatedMarkdown(Markdown):
    def __rich_console__(self, console, options):
        results = list(super().__rich_console__(console, options))
        height = console.height
        count = 0
        buffer = []
        for segment in reversed(results):
            count += segment.text.count("\n")  # type: ignore
            if count > height:
                break
            buffer.append(segment)

        self.truncated = len(buffer) < len(results)
        self.original = results

        yield from reversed(buffer)


@contextmanager
def streaming_markdown(get_md: Callable[[], str]):
    full_markdown: RenderResult | None = None

    @derived
    def get_ansi():
        segments = Console().render(m := TruncatedMarkdown(get_md()))
        ansi_output = "".join(seg.style.render(seg.text) if seg.style else seg.text for seg in segments)
        if m.truncated:
            nonlocal full_markdown
            full_markdown = m.original
        return ANSI(ansi_output.rstrip())

    event = Event()
    lock = Lock()

    app = Application(
        Layout(Window(FormattedTextControl(get_ansi), always_hide_cursor=True)),
        key_bindings=(kb := KeyBindings()),
        refresh_interval=0,  # disable automatic refresh
        after_render=lambda _: event.set(),
    )

    @async_effect(call_immediately=False, task_factory=lambda func: app.loop.create_task(func()))  # type: ignore
    async def update():
        async with lock:
            app.invalidate()
            await event.wait()
        event.clear()
        get_md()  # track dependency

    # manually connect dependencies and subscribers for the first time
    update.dependencies.add(get_ansi)
    get_ansi.subscribers.add(update)

    @kb.add("c-c")
    def _(_):
        raise KeyboardInterrupt

    thread = Thread(target=lambda: app.run(set_exception_handler=False, handle_sigint=False), daemon=True)
    thread.start()
    try:
        yield
    finally:
        if full_markdown:
            app.erase_when_done = True
        app.exit()
        thread.join()
        if full_markdown:
            Console().print(Renderables(full_markdown))



from contextlib import redirect_stderr
from functools import cache
from io import StringIO

from m.utils.path import global_store
from typer import Exit


@cache
def _get_session():
    from prompt_toolkit import PromptSession
    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.history import FileHistory
    from prompt_toolkit.input import create_input
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.keys import Keys
    from prompt_toolkit.styles import Style

    history = FileHistory(global_store / ".history")
    style = Style([("", "ansiyellow")])

    kb = KeyBindings()

    @kb.add(Keys.ControlJ)
    @kb.add(Keys.Escape, Keys.Enter)
    def _(event):
        event.current_buffer.insert_text("\n")

    with redirect_stderr(io := StringIO()):
        session = PromptSession(history=history, style=style, key_bindings=kb)

    if "Input is not a terminal (fd=0)" in io.getvalue():  # Warning: Input is not a terminal (fd=0).
        return PromptSession(input=create_input(open("/dev/tty")), history=history, style=style, key_bindings=kb)  # noqa: PTH123

    buf = session.default_buffer
    _accept = buf.accept_handler
    assert _accept is not None

    def accept(buf: Buffer):
        buf.text = buf.text.strip()
        return _accept(buf)

    buf.accept_handler = accept

    return session


def prompt():
    try:
        return _get_session().prompt(" > ", prompt_continuation="   ")
    finally:
        print()


def get_user_message():
    print()
    if message := prompt():
        return message
    else:
        raise Exit
